@@ -1,47 +1,59 @@
 /* copyright 2012 - 2013 Sascha Kruse and contributors (see LICENSE for licensing information) */
 
-#define XLIB_ILLEGAL_ACCESS
-
 #include "dunst.h"
 
-#include <X11/Xlib.h>
-#include <glib-unix.h>
+#include <assert.h>
 #include <glib.h>
+#include <glib-unix.h>
 #include <signal.h>
 #include <stdbool.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <X11/Xlib.h>
 
 #include "dbus.h"
+#include "draw.h"
+#include "log.h"
 #include "menu.h"
 #include "notification.h"
 #include "option_parser.h"
 #include "queues.h"
 #include "settings.h"
+#include "utils.h"
 #include "x11/screen.h"
 #include "x11/x.h"
 
-#ifndef VERSION
-#define VERSION "version info needed"
-#endif
-
-#define MSG 1
-#define INFO 2
-#define DEBUG 3
-
-typedef struct _x11_source {
-        GSource source;
-        Display *dpy;
-        Window w;
-} x11_source_t;
-
-/* index of colors fit to urgency level */
-
 GMainLoop *mainloop = NULL;
 
-GSList *rules = NULL;
+static struct dunst_status status;
 
-/* misc funtions */
+/* see dunst.h */
+void dunst_status(const enum dunst_status_field field,
+                  bool value)
+{
+        switch (field) {
+        case S_FULLSCREEN:
+                status.fullscreen = value;
+                break;
+        case S_IDLE:
+                status.idle = value;
+                break;
+        case S_RUNNING:
+                status.running = value;
+                break;
+        default:
+                LOG_E("Invalid %s enum value in %s:%d", "dunst_status", __FILE__, __LINE__);
+                break;
+        }
+}
+
+/* see dunst.h */
+struct dunst_status dunst_status_get(void)
+{
+        return status;
+}
+
+/* misc functions */
 static gboolean run(void *data);
 
 void wake_up(void)
@@ -51,25 +63,27 @@ void wake_up(void)
 
 static gboolean run(void *data)
 {
-        queues_check_timeouts(x_is_idle());
-        queues_update();
-
         static gint64 next_timeout = 0;
 
-        if (!xctx.visible && queues_length_displayed() > 0) {
-                x_win_show();
+        LOG_D("RUN");
+
+        dunst_status(S_FULLSCREEN, have_fullscreen_window());
+        dunst_status(S_IDLE, x_is_idle());
+
+        queues_update(status);
+
+        bool active = queues_length_displayed() > 0;
+
+        if (active) {
+                // Call draw before showing the window to avoid flickering
+                draw();
+                x_win_show(win);
+        } else {
+                x_win_hide(win);
         }
 
-        if (xctx.visible && queues_length_displayed() == 0) {
-                x_win_hide();
-        }
-
-        if (xctx.visible) {
-                x_win_draw();
-        }
-
-        if (xctx.visible) {
-                gint64 now = g_get_monotonic_time();
+        if (active) {
+                gint64 now = time_monotonic_now();
                 gint64 sleep = queues_get_next_datachange(now);
                 gint64 timeout_at = now + sleep;
 
@@ -91,7 +105,7 @@ static gboolean run(void *data)
 
 gboolean pause_signal(gpointer data)
 {
-        queues_pause_on();
+        dunst_status(S_RUNNING, false);
         wake_up();
 
         return G_SOURCE_CONTINUE;
@@ -99,7 +113,7 @@ gboolean pause_signal(gpointer data)
 
 gboolean unpause_signal(gpointer data)
 {
-        queues_pause_off();
+        dunst_status(S_RUNNING, true);
         wake_up();
 
         return G_SOURCE_CONTINUE;
@@ -116,23 +130,32 @@ static void teardown(void)
 {
         regex_teardown();
 
-        teardown_queues();
+        queues_teardown();
 
-        x_free();
+        draw_deinit();
 }
 
 int dunst_main(int argc, char *argv[])
 {
 
+        dunst_status(S_RUNNING, true);
+        dunst_status(S_IDLE, false);
+
         queues_init();
 
         cmdline_load(argc, argv);
 
+        dunst_log_init(false);
+
         if (cmdline_get_bool("-v/-version", false, "Print version")
             || cmdline_get_bool("--version", false, "Print version")) {
                 print_version();
         }
 
+        char *verbosity = cmdline_get_string("-verbosity", NULL, "Minimum level for message");
+        log_set_level_from_string(verbosity);
+        g_free(verbosity);
+
         char *cmdline_config_path;
         cmdline_config_path =
             cmdline_get_string("-conf/-config", NULL,
@@ -144,47 +167,11 @@ int dunst_main(int argc, char *argv[])
                 usage(EXIT_SUCCESS);
         }
 
-        int owner_id = initdbus();
-
-        x_setup();
-
-        if (settings.startup_notification) {
-                notification *n = notification_create();
-                n->id = 0;
-                n->appname = g_strdup("dunst");
-                n->summary = g_strdup("startup");
-                n->body = g_strdup("dunst is up and running");
-                n->progress = -1;
-                n->timeout = 10 * G_USEC_PER_SEC;
-                n->markup = MARKUP_NO;
-                n->urgency = URG_LOW;
-                notification_init(n);
-                queues_notification_insert(n);
-                // we do not call wakeup now, wake_up does not work here yet
-        }
+        int dbus_owner_id = dbus_init();
 
         mainloop = g_main_loop_new(NULL, FALSE);
 
-        GPollFD dpy_pollfd = { xctx.dpy->fd,
-                G_IO_IN | G_IO_HUP | G_IO_ERR, 0
-        };
-
-        GSourceFuncs x11_source_funcs = {
-                x_mainloop_fd_prepare,
-                x_mainloop_fd_check,
-                x_mainloop_fd_dispatch,
-                NULL,
-                NULL,
-                NULL
-        };
-
-        GSource *x11_source =
-            g_source_new(&x11_source_funcs, sizeof(x11_source_t));
-        ((x11_source_t *) x11_source)->dpy = xctx.dpy;
-        ((x11_source_t *) x11_source)->w = xctx.win;
-        g_source_add_poll(x11_source, &dpy_pollfd);
-
-        g_source_attach(x11_source, NULL);
+        draw_setup();
 
         guint pause_src = g_unix_signal_add(SIGUSR1, pause_signal, NULL);
         guint unpause_src = g_unix_signal_add(SIGUSR2, unpause_signal, NULL);
@@ -194,9 +181,24 @@ int dunst_main(int argc, char *argv[])
         guint term_src = g_unix_signal_add(SIGTERM, quit_signal, NULL);
         guint int_src = g_unix_signal_add(SIGINT, quit_signal, NULL);
 
+        if (settings.startup_notification) {
+                struct notification *n = notification_create();
+                n->id = 0;
+                n->appname = g_strdup("dunst");
+                n->summary = g_strdup("startup");
+                n->body = g_strdup("dunst is up and running");
+                n->progress = -1;
+                n->timeout = S2US(10);
+                n->markup = MARKUP_NO;
+                n->urgency = URG_LOW;
+                notification_init(n);
+                queues_notification_insert(n);
+                // we do not call wakeup now, wake_up does not work here yet
+        }
+
         run(NULL);
         g_main_loop_run(mainloop);
-        g_main_loop_unref(mainloop);
+        g_clear_pointer(&mainloop, g_main_loop_unref);
 
         /* remove signal handler watches */
         g_source_remove(pause_src);
@@ -204,9 +206,7 @@ int dunst_main(int argc, char *argv[])
         g_source_remove(term_src);
         g_source_remove(int_src);
 
-        g_source_destroy(x11_source);
-
-        dbus_tear_down(owner_id);
+        dbus_teardown(dbus_owner_id);
 
         teardown();
 
