@@ -15,38 +15,59 @@
 
 #include "dbus.h"
 #include "dunst.h"
+#include "icon.h"
+#include "log.h"
 #include "markup.h"
 #include "menu.h"
 #include "queues.h"
 #include "rules.h"
 #include "settings.h"
 #include "utils.h"
-#include "x11/x.h"
 
-static void notification_extract_urls(notification *n);
-static void notification_format_message(notification *n);
-static void notification_dmenu_string(notification *n);
+static void notification_extract_urls(struct notification *n);
+static void notification_format_message(struct notification *n);
 
-/*
- * print a human readable representation
- * of the given notification to stdout.
- */
-void notification_print(notification *n)
+/* see notification.h */
+const char *enum_to_string_fullscreen(enum behavior_fullscreen in)
 {
+        switch (in) {
+        case FS_SHOW: return "show";
+        case FS_DELAY: return "delay";
+        case FS_PUSHBACK: return "pushback";
+        case FS_NULL: return "(null)";
+        default:
+                LOG_E("Invalid %s enum value in %s:%d", "fullscreen", __FILE__, __LINE__);
+                break;
+        }
+}
+
+struct _notification_private {
+        gint refcount;
+};
+
+/* see notification.h */
+void notification_print(const struct notification *n)
+{
+        //TODO: use logging info for this
         printf("{\n");
         printf("\tappname: '%s'\n", n->appname);
         printf("\tsummary: '%s'\n", n->summary);
         printf("\tbody: '%s'\n", n->body);
-        printf("\ticon: '%s'\n", n->icon);
-        printf("\traw_icon set: %s\n", (n->raw_icon ? "true" : "false"));
+        printf("\ticon: '%s'\n", n->iconname);
+        printf("\traw_icon set: %s\n", (n->icon_id && !STR_EQ(n->iconname, n->icon_id)) ? "true" : "false");
+        printf("\ticon_id: '%s'\n", n->icon_id);
+        printf("\tdesktop_entry: '%s'\n", n->desktop_entry ? n->desktop_entry : "");
         printf("\tcategory: %s\n", n->category);
         printf("\ttimeout: %ld\n", n->timeout/1000);
         printf("\turgency: %s\n", notification_urgency_to_string(n->urgency));
         printf("\ttransient: %d\n", n->transient);
         printf("\tformatted: '%s'\n", n->msg);
-        printf("\tfg: %s\n", n->colors[ColFG]);
-        printf("\tbg: %s\n", n->colors[ColBG]);
-        printf("\tframe: %s\n", n->colors[ColFrame]);
+        printf("\tfg: %s\n", n->colors.fg);
+        printf("\tbg: %s\n", n->colors.bg);
+        printf("\tframe: %s\n", n->colors.frame);
+        printf("\tfullscreen: %s\n", enum_to_string_fullscreen(n->fullscreen));
+        printf("\tprogress: %d\n", n->progress);
+        printf("\tstack_tag: %s\n", (n->stack_tag ? n->stack_tag : ""));
         printf("\tid: %d\n", n->id);
         if (n->urls) {
                 char *urls = string_replace_all("\n", "\t\t\n", g_strdup(n->urls));
@@ -56,34 +77,36 @@ void notification_print(notification *n)
                 printf("\t}\n");
                 g_free(urls);
         }
-
-        if (n->actions) {
-                printf("\tactions:\n");
-                printf("\t{\n");
-                for (int i = 0; i < n->actions->count; i += 2) {
-                        printf("\t\t[%s,%s]\n", n->actions->actions[i],
-                               n->actions->actions[i + 1]);
-                }
+        if (g_hash_table_size(n->actions) == 0) {
+                printf("\tactions: {}\n");
+        } else {
+                gpointer p_key, p_value;
+                GHashTableIter iter;
+                g_hash_table_iter_init(&iter, n->actions);
+                printf("\tactions: {\n");
+                while (g_hash_table_iter_next(&iter, &p_key, &p_value))
+                        printf("\t\t\"%s\": \"%s\"\n", (char*)p_key, (char*)p_value);
                 printf("\t}\n");
-                printf("\tactions_dmenu: %s\n", n->actions->dmenu_str);
         }
         printf("\tscript: %s\n", n->script);
         printf("}\n");
 }
 
-/*
- * Run the script associated with the
- * given notification.
- */
-void notification_run_script(notification *n)
+/* see notification.h */
+void notification_run_script(struct notification *n)
 {
-        if (!n->script || strlen(n->script) < 1)
+        if (STR_EMPTY(n->script))
                 return;
 
-        char *appname = n->appname ? n->appname : "";
-        char *summary = n->summary ? n->summary : "";
-        char *body = n->body ? n->body : "";
-        char *icon = n->icon ? n->icon : "";
+        if (n->script_run && !settings.always_run_script)
+                return;
+
+        n->script_run = true;
+
+        const char *appname = n->appname ? n->appname : "";
+        const char *summary = n->summary ? n->summary : "";
+        const char *body = n->body ? n->body : "";
+        const char *icon = n->iconname ? n->iconname : "";
 
         const char *urgency = notification_urgency_to_string(n->urgency);
 
@@ -106,7 +129,7 @@ void notification_run_script(notification *n)
                                          urgency,
                                          (char *)NULL);
                         if (ret != 0) {
-                                PERR("Unable to run script", errno);
+                                LOG_W("Unable to run script: %s", strerror(errno));
                                 exit(EXIT_FAILURE);
                         }
                 }
@@ -116,7 +139,7 @@ void notification_run_script(notification *n)
 /*
  * Helper function to convert an urgency to a string
  */
-const char *notification_urgency_to_string(enum urgency urgency)
+const char *notification_urgency_to_string(const enum urgency urgency)
 {
         switch (urgency) {
         case URG_NONE:
@@ -132,18 +155,9 @@ const char *notification_urgency_to_string(enum urgency urgency)
         }
 }
 
-/*
- * Helper function to compare to given
- * notifications.
- */
-int notification_cmp(const void *va, const void *vb)
+/* see notification.h */
+int notification_cmp(const struct notification *a, const struct notification *b)
 {
-        notification *a = (notification *) va;
-        notification *b = (notification *) vb;
-
-        if (!settings.sort)
-                return 1;
-
         if (a->urgency != b->urgency) {
                 return b->urgency - a->urgency;
         } else {
@@ -151,89 +165,109 @@ int notification_cmp(const void *va, const void *vb)
         }
 }
 
-/*
- * Wrapper for notification_cmp to match glib's
- * compare functions signature.
- */
+/* see notification.h */
 int notification_cmp_data(const void *va, const void *vb, void *data)
 {
-        return notification_cmp(va, vb);
+        struct notification *a = (struct notification *) va;
+        struct notification *b = (struct notification *) vb;
+
+        ASSERT_OR_RET(settings.sort, 1);
+
+        return notification_cmp(a, b);
 }
 
-int notification_is_duplicate(const notification *a, const notification *b)
+bool notification_is_duplicate(const struct notification *a, const struct notification *b)
 {
-        //Comparing raw icons is not supported, assume they are not identical
-        if (settings.icon_position != icons_off
-                && (a->raw_icon != NULL || b->raw_icon != NULL))
-                return false;
-
-        return strcmp(a->appname, b->appname) == 0
-            && strcmp(a->summary, b->summary) == 0
-            && strcmp(a->body,    b->body) == 0
-            && (settings.icon_position != icons_off ? strcmp(a->icon, b->icon) == 0 : 1)
+        return STR_EQ(a->appname, b->appname)
+            && STR_EQ(a->summary, b->summary)
+            && STR_EQ(a->body, b->body)
+            && (settings.icon_position != ICON_OFF ? STR_EQ(a->icon_id, b->icon_id) : 1)
             && a->urgency == b->urgency;
 }
 
-/*
- * Free the actions element
- * @a: (nullable): Pointer to #Actions
- */
-void actions_free(Actions *a)
+static void notification_private_free(NotificationPrivate *p)
 {
-        if (!a)
-                return;
-
-        g_strfreev(a->actions);
-        g_free(a->dmenu_str);
-        g_free(a);
+        g_free(p);
 }
 
-/*
- * Free a #RawImage
- * @i: (nullable): pointer to #RawImage
- */
-void rawimage_free(RawImage *i)
+/* see notification.h */
+gint notification_refcount_get(struct notification *n)
 {
-        if (!i)
-                return;
-
-        g_free(i->data);
-        g_free(i);
+        assert(n->priv->refcount > 0);
+        return g_atomic_int_get(&n->priv->refcount);
 }
 
-/*
- * Free the memory used by the given notification.
- */
-void notification_free(notification *n)
+/* see notification.h */
+void notification_ref(struct notification *n)
 {
-        assert(n != NULL);
+        assert(n->priv->refcount > 0);
+        g_atomic_int_inc(&n->priv->refcount);
+}
+
+/* see notification.h */
+void notification_unref(struct notification *n)
+{
+        ASSERT_OR_RET(n,);
+
+        assert(n->priv->refcount > 0);
+        if (!g_atomic_int_dec_and_test(&n->priv->refcount))
+                return;
+
         g_free(n->appname);
         g_free(n->summary);
         g_free(n->body);
-        g_free(n->icon);
+        g_free(n->iconname);
         g_free(n->msg);
         g_free(n->dbus_client);
         g_free(n->category);
         g_free(n->text_to_render);
         g_free(n->urls);
-        g_free(n->colors[ColFG]);
-        g_free(n->colors[ColBG]);
-        g_free(n->colors[ColFrame]);
+        g_free(n->colors.fg);
+        g_free(n->colors.bg);
+        g_free(n->colors.frame);
+        g_free(n->stack_tag);
+        g_free(n->desktop_entry);
 
-        actions_free(n->actions);
-        rawimage_free(n->raw_icon);
+        g_hash_table_unref(n->actions);
+
+        if (n->icon)
+                g_object_unref(n->icon);
+        g_free(n->icon_id);
+
+        notification_private_free(n->priv);
 
         g_free(n);
 }
 
-/*
- * Replace the two chars where **needle points
- * with a quoted "replacement", according to the markup settings.
- *
- * The needle is a double pointer and gets updated upon return
- * to point to the first char, which occurs after replacement.
- *
- */
+void notification_icon_replace_path(struct notification *n, const char *new_icon)
+{
+        ASSERT_OR_RET(n,);
+        ASSERT_OR_RET(new_icon,);
+        ASSERT_OR_RET(n->iconname != new_icon,);
+
+        g_free(n->iconname);
+        n->iconname = g_strdup(new_icon);
+
+        g_clear_object(&n->icon);
+        g_clear_pointer(&n->icon_id, g_free);
+
+        n->icon = icon_get_for_name(new_icon, &n->icon_id);
+        n->icon = icon_pixbuf_scale(n->icon);
+}
+
+void notification_icon_replace_data(struct notification *n, GVariant *new_icon)
+{
+        ASSERT_OR_RET(n,);
+        ASSERT_OR_RET(new_icon,);
+
+        g_clear_object(&n->icon);
+        g_clear_pointer(&n->icon_id, g_free);
+
+        n->icon = icon_get_for_data(new_icon, &n->icon_id);
+        n->icon = icon_pixbuf_scale(n->icon);
+}
+
+/* see notification.h */
 void notification_replace_single_field(char **haystack,
                                        char **needle,
                                        const char *replacement,
@@ -257,24 +291,27 @@ void notification_replace_single_field(char **haystack,
         g_free(input);
 }
 
-/*
- * Create notification struct and initialise all fields with either
- *  - the default (if it's not needed to be freed later)
- *  - its undefined representation (NULL, -1)
- *
- * This function is guaranteed to return a valid pointer.
- * @Returns: The generated notification
- */
-notification *notification_create(void)
+static NotificationPrivate *notification_private_create(void)
 {
-        notification *n = g_malloc0(sizeof(notification));
+        NotificationPrivate *priv = g_malloc0(sizeof(NotificationPrivate));
+        g_atomic_int_set(&priv->refcount, 1);
+
+        return priv;
+}
+
+/* see notification.h */
+struct notification *notification_create(void)
+{
+        struct notification *n = g_malloc0(sizeof(struct notification));
+
+        n->priv = notification_private_create();
 
         /* Unparameterized default values */
         n->first_render = true;
         n->markup = settings.markup;
         n->format = settings.format;
 
-        n->timestamp = g_get_monotonic_time();
+        n->timestamp = time_monotonic_now();
 
         n->urgency = URG_NORM;
         n->timeout = -1;
@@ -282,16 +319,18 @@ notification *notification_create(void)
         n->transient = false;
         n->progress = -1;
 
+        n->script_run = false;
+        n->dbus_valid = false;
+
+        n->fullscreen = FS_SHOW;
+
+        n->actions = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, g_free);
+
         return n;
 }
 
-/*
- * Sanitize values of notification, apply all matching rules
- * and generate derived fields.
- *
- * @n: the notification to sanitize
- */
-void notification_init(notification *n)
+/* see notification.h */
+void notification_init(struct notification *n)
 {
         /* default to empty string to avoid further NULL faults */
         n->appname  = n->appname  ? n->appname  : g_strdup("unknown");
@@ -310,21 +349,40 @@ void notification_init(notification *n)
                 n->timeout = settings.timeouts[n->urgency];
 
         /* Icon handling */
-        if (n->icon && strlen(n->icon) <= 0)
-                g_clear_pointer(&n->icon, g_free);
-        if (!n->raw_icon && !n->icon)
-                n->icon = g_strdup(settings.icons[n->urgency]);
+        if (STR_EMPTY(n->iconname))
+                g_clear_pointer(&n->iconname, g_free);
+        if (!n->icon && n->iconname) {
+                char *icon = g_strdup(n->iconname);
+                notification_icon_replace_path(n, icon);
+                g_free(icon);
+        }
+        if (!n->icon && !n->iconname)
+                notification_icon_replace_path(n, settings.icons[n->urgency]);
 
         /* Color hints */
-        if (!n->colors[ColFG])
-                n->colors[ColFG] = g_strdup(xctx.colors[ColFG][n->urgency]);
-        if (!n->colors[ColBG])
-                n->colors[ColBG] = g_strdup(xctx.colors[ColBG][n->urgency]);
-        if (!n->colors[ColFrame])
-                n->colors[ColFrame] = g_strdup(xctx.colors[ColFrame][n->urgency]);
+        struct notification_colors defcolors;
+        switch (n->urgency) {
+                case URG_LOW:
+                        defcolors = settings.colors_low;
+                        break;
+                case URG_NORM:
+                        defcolors = settings.colors_norm;
+                        break;
+                case URG_CRIT:
+                        defcolors = settings.colors_crit;
+                        break;
+                default:
+                        g_error("Unhandled urgency type: %d", n->urgency);
+        }
+        if (!n->colors.fg)
+                n->colors.fg = g_strdup(defcolors.fg);
+        if (!n->colors.bg)
+                n->colors.bg = g_strdup(defcolors.bg);
+        if (!n->colors.frame)
+                n->colors.frame = g_strdup(defcolors.frame);
 
         /* Sanitize misc hints */
-        if (n->progress < 0 || n->progress > 100)
+        if (n->progress < 0)
                 n->progress = -1;
 
         /* Process rules */
@@ -332,11 +390,10 @@ void notification_init(notification *n)
 
         /* UPDATE derived fields */
         notification_extract_urls(n);
-        notification_dmenu_string(n);
         notification_format_message(n);
 }
 
-static void notification_format_message(notification *n)
+static void notification_format_message(struct notification *n)
 {
         g_clear_pointer(&n->msg, g_free);
 
@@ -344,7 +401,7 @@ static void notification_format_message(notification *n)
 
         /* replace all formatter */
         for(char *substr = strchr(n->msg, '%');
-                  substr;
+                  substr && *substr;
                   substr = strchr(substr, '%')) {
 
                 char pg[16];
@@ -363,7 +420,7 @@ static void notification_format_message(notification *n)
                                 &n->msg,
                                 &substr,
                                 n->summary,
-                                n->markup);
+                                MARKUP_NO);
                         break;
                 case 'b':
                         notification_replace_single_field(
@@ -373,7 +430,7 @@ static void notification_format_message(notification *n)
                                 n->markup);
                         break;
                 case 'I':
-                        icon_tmp = g_strdup(n->icon);
+                        icon_tmp = g_strdup(n->iconname);
                         notification_replace_single_field(
                                 &n->msg,
                                 &substr,
@@ -385,7 +442,7 @@ static void notification_format_message(notification *n)
                         notification_replace_single_field(
                                 &n->msg,
                                 &substr,
-                                n->icon ? n->icon : "",
+                                n->iconname ? n->iconname : "",
                                 MARKUP_NO);
                         break;
                 case 'p':
@@ -416,12 +473,12 @@ static void notification_format_message(notification *n)
                                 MARKUP_NO);
                         break;
                 case '\0':
-                        fprintf(stderr, "WARNING: format_string has trailing %% character."
-                                        "To escape it use %%%%.");
+                        LOG_W("format_string has trailing %% character. "
+                              "To escape it use %%%%.");
+                        substr++;
                         break;
                 default:
-                        fprintf(stderr, "WARNING: format_string %%%c"
-                                        " is unknown\n", substr[1]);
+                        LOG_W("format_string %%%c is unknown.", substr[1]);
                         // shift substr pointer forward,
                         // as we can't interpret the format string
                         substr++;
@@ -432,17 +489,14 @@ static void notification_format_message(notification *n)
         n->msg = g_strchomp(n->msg);
 
         /* truncate overlong messages */
-        if (strlen(n->msg) > DUNST_NOTIF_MAX_CHARS) {
-                char *buffer = g_malloc(DUNST_NOTIF_MAX_CHARS);
-                strncpy(buffer, n->msg, DUNST_NOTIF_MAX_CHARS);
-                buffer[DUNST_NOTIF_MAX_CHARS-1] = '\0';
-
+        if (strnlen(n->msg, DUNST_NOTIF_MAX_CHARS + 1) > DUNST_NOTIF_MAX_CHARS) {
+                char * buffer = g_strndup(n->msg, DUNST_NOTIF_MAX_CHARS);
                 g_free(n->msg);
                 n->msg = buffer;
         }
 }
 
-static void notification_extract_urls(notification *n)
+static void notification_extract_urls(struct notification *n)
 {
         g_clear_pointer(&n->urls, g_free);
 
@@ -466,25 +520,8 @@ static void notification_extract_urls(notification *n)
         g_free(urls_text);
 }
 
-static void notification_dmenu_string(notification *n)
-{
-        if (n->actions) {
-                g_clear_pointer(&n->actions->dmenu_str, g_free);
-                for (int i = 0; i < n->actions->count; i += 2) {
-                        char *human_readable = n->actions->actions[i + 1];
-                        string_replace_char('[', '(', human_readable); // kill square brackets
-                        string_replace_char(']', ')', human_readable);
 
-                        char *act_str = g_strdup_printf("#%s [%s]", human_readable, n->appname);
-                        if (act_str) {
-                                n->actions->dmenu_str = string_append(n->actions->dmenu_str, act_str, "\n");
-                                g_free(act_str);
-                        }
-                }
-        }
-}
-
-void notification_update_text_to_render(notification *n)
+void notification_update_text_to_render(struct notification *n)
 {
         g_clear_pointer(&n->text_to_render, g_free);
 
@@ -494,14 +531,14 @@ void notification_update_text_to_render(notification *n)
 
         /* print dup_count and msg */
         if ((n->dup_count > 0 && !settings.hide_duplicate_count)
-            && (n->actions || n->urls) && settings.show_indicators) {
+            && (g_hash_table_size(n->actions) || n->urls) && settings.show_indicators) {
                 buf = g_strdup_printf("(%d%s%s) %s",
                                       n->dup_count,
-                                      n->actions ? "A" : "",
+                                      g_hash_table_size(n->actions) ? "A" : "",
                                       n->urls ? "U" : "", msg);
-        } else if ((n->actions || n->urls) && settings.show_indicators) {
+        } else if ((g_hash_table_size(n->actions) || n->urls) && settings.show_indicators) {
                 buf = g_strdup_printf("(%s%s) %s",
-                                      n->actions ? "A" : "",
+                                      g_hash_table_size(n->actions) ? "A" : "",
                                       n->urls ? "U" : "", msg);
         } else if (n->dup_count > 0 && !settings.hide_duplicate_count) {
                 buf = g_strdup_printf("(%d) %s", n->dup_count, msg);
@@ -511,7 +548,7 @@ void notification_update_text_to_render(notification *n)
 
         /* print age */
         gint64 hours, minutes, seconds;
-        gint64 t_delta = g_get_monotonic_time() - n->timestamp;
+        gint64 t_delta = time_monotonic_now() - n->timestamp;
 
         if (settings.show_age_threshold >= 0
             && t_delta >= settings.show_age_threshold) {
@@ -539,31 +576,27 @@ void notification_update_text_to_render(notification *n)
         n->text_to_render = buf;
 }
 
-/*
- * If the notification has exactly one action, or one is marked as default,
- * invoke it. If there are multiple and no default, open the context menu. If
- * there are no actions, proceed similarly with urls.
- */
-void notification_do_action(notification *n)
+/* see notification.h */
+void notification_do_action(const struct notification *n)
 {
-        if (n->actions) {
-                if (n->actions->count == 2) {
-                        signal_action_invoked(n, n->actions->actions[0]);
+        if (g_hash_table_size(n->actions)) {
+                if (g_hash_table_contains(n->actions, "default")) {
+                        signal_action_invoked(n, "default");
                         return;
                 }
-                for (int i = 0; i < n->actions->count; i += 2) {
-                        if (strcmp(n->actions->actions[i], "default") == 0) {
-                                signal_action_invoked(n, n->actions->actions[i]);
-                                return;
-                        }
+                if (g_hash_table_size(n->actions) == 1) {
+                        GList *keys = g_hash_table_get_keys(n->actions);
+                        signal_action_invoked(n, keys->data);
+                        g_list_free(keys);
+                        return;
                 }
                 context_menu();
 
         } else if (n->urls) {
-                if (strstr(n->urls, "\n") == NULL)
-                        open_browser(n->urls);
-                else
+                if (strstr(n->urls, "\n"))
                         context_menu();
+                else
+                        open_browser(n->urls);
         }
 }
 
