@@ -8,27 +8,29 @@
 #include <stdlib.h>
 #include <string.h>
 
+#include "dunst.h"
+#include "log.h"
 #include "utils.h"
+#include "settings.h"
 
-typedef struct _entry_t {
+struct entry {
         char *key;
         char *value;
-} entry_t;
+};
 
-typedef struct _section_t {
+struct section {
         char *name;
         int entry_count;
-        entry_t *entries;
-} section_t;
+        struct entry *entries;
+};
 
 static int section_count = 0;
-static section_t *sections;
+static struct section *sections;
 
-static section_t *new_section(const char *name);
-static section_t *get_section(const char *name);
+static struct section *new_section(const char *name);
+static struct section *get_section(const char *name);
 static void add_entry(const char *section_name, const char *key, const char *value);
 static const char *get_value(const char *section, const char *key);
-static char *clean_value(const char *value);
 
 static int cmdline_argc;
 static char **cmdline_argv;
@@ -38,16 +40,132 @@ static void cmdline_usage_append(const char *key, const char *type, const char *
 
 static int cmdline_find_option(const char *key);
 
-section_t *new_section(const char *name)
+#define STRING_PARSE_RET(string, value) if (STR_EQ(s, string)) { *ret = value; return true; }
+
+bool string_parse_alignment(const char *s, enum alignment *ret)
+{
+        ASSERT_OR_RET(s, false);
+        ASSERT_OR_RET(ret, false);
+
+        STRING_PARSE_RET("left",   ALIGN_LEFT);
+        STRING_PARSE_RET("center", ALIGN_CENTER);
+        STRING_PARSE_RET("right",  ALIGN_RIGHT);
+
+        return false;
+}
+
+bool string_parse_ellipsize(const char *s, enum ellipsize *ret)
+{
+        ASSERT_OR_RET(s, false);
+        ASSERT_OR_RET(ret, false);
+
+        STRING_PARSE_RET("start",  ELLIPSE_START);
+        STRING_PARSE_RET("middle", ELLIPSE_MIDDLE);
+        STRING_PARSE_RET("end",    ELLIPSE_END);
+
+        return false;
+}
+
+bool string_parse_follow_mode(const char *s, enum follow_mode *ret)
+{
+        ASSERT_OR_RET(s, false);
+        ASSERT_OR_RET(ret, false);
+
+        STRING_PARSE_RET("mouse",    FOLLOW_MOUSE);
+        STRING_PARSE_RET("keyboard", FOLLOW_KEYBOARD);
+        STRING_PARSE_RET("none",     FOLLOW_NONE);
+
+        return false;
+}
+
+
+bool string_parse_fullscreen(const char *s, enum behavior_fullscreen *ret)
+{
+        ASSERT_OR_RET(s, false);
+        ASSERT_OR_RET(ret, false);
+
+        STRING_PARSE_RET("show",     FS_SHOW);
+        STRING_PARSE_RET("delay",    FS_DELAY);
+        STRING_PARSE_RET("pushback", FS_PUSHBACK);
+
+        return false;
+}
+
+bool string_parse_icon_position(const char *s, enum icon_position *ret)
+{
+        ASSERT_OR_RET(s, false);
+        ASSERT_OR_RET(ret, false);
+
+        STRING_PARSE_RET("left",  ICON_LEFT);
+        STRING_PARSE_RET("right", ICON_RIGHT);
+        STRING_PARSE_RET("off",   ICON_OFF);
+
+        return false;
+}
+
+bool string_parse_markup_mode(const char *s, enum markup_mode *ret)
+{
+        ASSERT_OR_RET(s, false);
+        ASSERT_OR_RET(ret, false);
+
+        STRING_PARSE_RET("strip", MARKUP_STRIP);
+        STRING_PARSE_RET("no",    MARKUP_NO);
+        STRING_PARSE_RET("full",  MARKUP_FULL);
+        STRING_PARSE_RET("yes",   MARKUP_FULL);
+
+        return false;
+}
+
+bool string_parse_mouse_action(const char *s, enum mouse_action *ret)
+{
+        ASSERT_OR_RET(s, false);
+        ASSERT_OR_RET(ret, false);
+
+        STRING_PARSE_RET("none",           MOUSE_NONE);
+        STRING_PARSE_RET("do_action",      MOUSE_DO_ACTION);
+        STRING_PARSE_RET("close_current",  MOUSE_CLOSE_CURRENT);
+        STRING_PARSE_RET("close_all",      MOUSE_CLOSE_ALL);
+
+        return false;
+}
+
+bool string_parse_sepcolor(const char *s, struct separator_color_data *ret)
+{
+        ASSERT_OR_RET(s, false);
+        ASSERT_OR_RET(ret, false);
+
+        STRING_PARSE_RET("auto",       (struct separator_color_data){.type = SEP_AUTO});
+        STRING_PARSE_RET("foreground", (struct separator_color_data){.type = SEP_FOREGROUND});
+        STRING_PARSE_RET("frame",      (struct separator_color_data){.type = SEP_FRAME});
+
+        ret->type = SEP_CUSTOM;
+        ret->sep_color = g_strdup(s);
+
+        return true;
+}
+
+bool string_parse_urgency(const char *s, enum urgency *ret)
+{
+        ASSERT_OR_RET(s, false);
+        ASSERT_OR_RET(ret, false);
+
+        STRING_PARSE_RET("low",      URG_LOW);
+        STRING_PARSE_RET("normal",   URG_NORM);
+        STRING_PARSE_RET("critical", URG_CRIT);
+
+        return false;
+}
+
+struct section *new_section(const char *name)
 {
         for (int i = 0; i < section_count; i++) {
-                if (!strcmp(name, sections[i].name)) {
-                        die("Duplicated section in dunstrc detected.\n", -1);
+                if (STR_EQ(name, sections[i].name)) {
+                        DIE("Duplicated section in dunstrc detected.");
                 }
         }
 
         section_count++;
-        sections = g_realloc(sections, sizeof(section_t) * section_count);
+        sections = g_realloc(sections, sizeof(struct section) * section_count);
         sections[section_count - 1].name = g_strdup(name);
         sections[section_count - 1].entries = NULL;
         sections[section_count - 1].entry_count = 0;
@@ -64,15 +182,14 @@ void free_ini(void)
                 g_free(sections[i].entries);
                 g_free(sections[i].name);
         }
-        g_free(sections);
+        g_clear_pointer(&sections, g_free);
         section_count = 0;
-        sections = NULL;
 }
 
-section_t *get_section(const char *name)
+struct section *get_section(const char *name)
 {
         for (int i = 0; i < section_count; i++) {
-                if (strcmp(sections[i].name, name) == 0)
+                if (STR_EQ(sections[i].name, name))
                         return &sections[i];
         }
 
@@ -81,27 +198,24 @@ section_t *get_section(const char *name)
 
 void add_entry(const char *section_name, const char *key, const char *value)
 {
-        section_t *s = get_section(section_name);
-        if (s == NULL) {
+        struct section *s = get_section(section_name);
+        if (!s)
                 s = new_section(section_name);
-        }
 
         s->entry_count++;
         int len = s->entry_count;
-        s->entries = g_realloc(s->entries, sizeof(entry_t) * len);
+        s->entries = g_realloc(s->entries, sizeof(struct entry) * len);
         s->entries[s->entry_count - 1].key = g_strdup(key);
-        s->entries[s->entry_count - 1].value = clean_value(value);
+        s->entries[s->entry_count - 1].value = string_strip_quotes(value);
 }
 
 const char *get_value(const char *section, const char *key)
 {
-        section_t *s = get_section(section);
-        if (!s) {
-                return NULL;
-        }
+        struct section *s = get_section(section);
+        ASSERT_OR_RET(s, NULL);
 
         for (int i = 0; i < s->entry_count; i++) {
-                if (strcmp(s->entries[i].key, key) == 0) {
+                if (STR_EQ(s->entries[i].key, key)) {
                         return s->entries[i].value;
                 }
         }
@@ -137,19 +251,19 @@ gint64 ini_get_time(const char *section, const char *key, gint64 def)
 int ini_get_int(const char *section, const char *key, int def)
 {
         const char *value = get_value(section, key);
-        if (value == NULL)
-                return def;
-        else
+        if (value)
                 return atoi(value);
+        else
+                return def;
 }
 
 double ini_get_double(const char *section, const char *key, double def)
 {
         const char *value = get_value(section, key);
-        if (value == NULL)
-                return def;
-        else
+        if (value)
                 return atof(value);
+        else
+                return def;
 }
 
 bool ini_is_set(const char *ini_section, const char *ini_key)
@@ -159,15 +273,11 @@ bool ini_is_set(const char *ini_section, const char *ini_key)
 
 const char *next_section(const char *section)
 {
-        if (section_count == 0)
-                return NULL;
-
-        if (section == NULL) {
-                return sections[0].name;
-        }
+        ASSERT_OR_RET(section_count > 0, NULL);
+        ASSERT_OR_RET(section, sections[0].name);
 
         for (int i = 0; i < section_count; i++) {
-                if (strcmp(section, sections[i].name) == 0) {
+                if (STR_EQ(section, sections[i].name)) {
                         if (i + 1 >= section_count)
                                 return NULL;
                         else
@@ -180,9 +290,7 @@ const char *next_section(const char *section)
 int ini_get_bool(const char *section, const char *key, int def)
 {
         const char *value = get_value(section, key);
-        if (value == NULL)
-                return def;
-        else {
+        if (value) {
                 switch (value[0]) {
                 case 'y':
                 case 'Y':
@@ -199,28 +307,14 @@ int ini_get_bool(const char *section, const char *key, int def)
                 default:
                         return def;
                 }
+        } else {
+                return def;
         }
 }
 
-char *clean_value(const char *value)
-{
-        char *s;
-
-        if (value[0] == '"')
-                s = g_strdup(value + 1);
-        else
-                s = g_strdup(value);
-
-        if (s[strlen(s) - 1] == '"')
-                s[strlen(s) - 1] = '\0';
-
-        return s;
-}
-
 int load_ini_file(FILE *fp)
 {
-        if (!fp)
-                return 1;
+        ASSERT_OR_RET(fp, 1);
 
         char *line = NULL;
         size_t line_len = 0;
@@ -232,16 +326,13 @@ int load_ini_file(FILE *fp)
 
                 char *start = g_strstrip(line);
 
-                if (*start == ';' || *start == '#' || strlen(start) == 0)
+                if (*start == ';' || *start == '#' || STR_EMPTY(start))
                         continue;
 
                 if (*start == '[') {
                         char *end = strchr(start + 1, ']');
                         if (!end) {
-                                fprintf(stderr,
-                                     "Warning: invalid config file at line %d\n",
-                                     line_num);
-                                fprintf(stderr, "Missing ']'\n");
+                                LOG_W("Invalid config file at line %d: Missing ']'.", line_num);
                                 continue;
                         }
 
@@ -255,10 +346,7 @@ int load_ini_file(FILE *fp)
 
                 char *equal = strchr(start + 1, '=');
                 if (!equal) {
-                        fprintf(stderr,
-                               "Warning: invalid config file at line %d\n",
-                               line_num);
-                        fprintf(stderr, "Missing '='\n");
+                        LOG_W("Invalid config file at line %d: Missing '='.", line_num);
                         continue;
                 }
 
@@ -267,27 +355,25 @@ int load_ini_file(FILE *fp)
                 char *value = g_strstrip(equal + 1);
 
                 char *quote = strchr(value, '"');
+                char *value_end = NULL;
                 if (quote) {
-                        char *closing_quote = strchr(quote + 1, '"');
-                        if (!closing_quote) {
-                                fprintf(stderr,
-                                     "Warning: invalid config file at line %d\n",
-                                     line_num);
-                                fprintf(stderr, "Missing '\"'\n");
+                        value_end = strchr(quote + 1, '"');
+                        if (!value_end) {
+                                LOG_W("Invalid config file at line %d: Missing '\"'.", line_num);
                                 continue;
                         }
                 } else {
-                        char *comment = strpbrk(value, "#;");
-                        if (comment)
-                                *comment = '\0';
+                        value_end = value;
                 }
+
+                char *comment = strpbrk(value_end, "#;");
+                if (comment)
+                        *comment = '\0';
+
                 value = g_strstrip(value);
 
                 if (!current_section) {
-                        fprintf(stderr,
-                               "Warning: invalid config file at line %d\n",
-                               line_num);
-                        fprintf(stderr, "Key value pair without a section\n");
+                        LOG_W("Invalid config file at line %d: Key value pair without a section.", line_num);
                         continue;
                 }
 
@@ -306,9 +392,8 @@ void cmdline_load(int argc, char *argv[])
 
 int cmdline_find_option(const char *key)
 {
-        if (!key) {
-                return -1;
-        }
+        ASSERT_OR_RET(key, -1);
+
         char *key1 = g_strdup(key);
         char *key2 = strchr(key1, '/');
 
@@ -319,7 +404,7 @@ int cmdline_find_option(const char *key)
 
         /* look for first key */
         for (int i = 0; i < cmdline_argc; i++) {
-                if (strcmp(key1, cmdline_argv[i]) == 0) {
+                if (STR_EQ(key1, cmdline_argv[i])) {
                         g_free(key1);
                         return i;
                 }
@@ -328,7 +413,7 @@ int cmdline_find_option(const char *key)
         /* look for second key if one was specified */
         if (key2) {
                 for (int i = 0; i < cmdline_argc; i++) {
-                        if (strcmp(key2, cmdline_argv[i]) == 0) {
+                        if (STR_EQ(key2, cmdline_argv[i])) {
                                 g_free(key1);
                                 return i;
                         }
@@ -348,8 +433,7 @@ static const char *cmdline_get_value(const char *key)
 
         if (idx + 1 >= cmdline_argc) {
                 /* the argument is missing */
-                fprintf(stderr, "Warning: %s, missing argument. Ignoring\n",
-                        key);
+                LOG_W("%s: Missing argument. Ignoring.", key);
                 return NULL;
         }
         return cmdline_argv[idx + 1];
@@ -362,10 +446,10 @@ char *cmdline_get_string(const char *key, const char *def, const char *descripti
 
         if (str)
                 return g_strdup(str);
-        if (def == NULL)
-                return NULL;
-        else
+        if (def)
                 return g_strdup(def);
+        else
+                return NULL;
 }
 
 char *cmdline_get_path(const char *key, const char *def, const char *description)
@@ -397,10 +481,10 @@ int cmdline_get_int(const char *key, int def, const char *description)
         cmdline_usage_append(key, "int", description);
         const char *str = cmdline_get_value(key);
 
-        if (str == NULL)
-                return def;
-        else
+        if (str)
                 return atoi(str);
+        else
+                return def;
 }
 
 double cmdline_get_double(const char *key, double def, const char *description)
@@ -408,10 +492,10 @@ double cmdline_get_double(const char *key, double def, const char *description)
         cmdline_usage_append(key, "double", description);
         const char *str = cmdline_get_value(key);
 
-        if (str == NULL)
-                return def;
-        else
+        if (str)
                 return atof(str);
+        else
+                return def;
 }
 
 int cmdline_get_bool(const char *key, int def, const char *description)
@@ -535,7 +619,7 @@ int option_get_bool(const char *ini_section,
 void cmdline_usage_append(const char *key, const char *type, const char *description)
 {
         char *key_type;
-        if (type && strlen(type) > 0)
+        if (STR_FULL(type))
                 key_type = g_strdup_printf("%s (%s)", key, type);
         else
                 key_type = g_strdup(key);
