@@ -6,6 +6,19 @@
 #include <stdbool.h>
 #include <stdio.h>
 
+#include "dunst.h"
+#include "settings.h"
+
+bool string_parse_alignment(const char *s, enum alignment *ret);
+bool string_parse_ellipsize(const char *s, enum ellipsize *ret);
+bool string_parse_follow_mode(const char *s, enum follow_mode *ret);
+bool string_parse_fullscreen(const char *s, enum behavior_fullscreen *ret);
+bool string_parse_icon_position(const char *s, enum icon_position *ret);
+bool string_parse_markup_mode(const char *s, enum markup_mode *ret);
+bool string_parse_mouse_action(const char *s, enum mouse_action *ret);
+bool string_parse_sepcolor(const char *s, struct separator_color_data *ret);
+bool string_parse_urgency(const char *s, enum urgency *ret);
+
 int load_ini_file(FILE *);
 char *ini_get_path(const char *section, const char *key, const char *def);
 char *ini_get_string(const char *section, const char *key, const char *def);
