@@ -4,36 +4,48 @@
 
 #include <stdbool.h>
 
+#include "markup.h"
+#include "notification.h"
 #include "x11/x.h"
 
-enum alignment { left, center, right };
-enum ellipsize { start, middle, end };
-enum icon_position_t { icons_left, icons_right, icons_off };
-enum separator_color { FOREGROUND, AUTO, FRAME, CUSTOM };
+enum alignment { ALIGN_LEFT, ALIGN_CENTER, ALIGN_RIGHT };
+enum ellipsize { ELLIPSE_START, ELLIPSE_MIDDLE, ELLIPSE_END };
+enum icon_position { ICON_LEFT, ICON_RIGHT, ICON_OFF };
+enum separator_color { SEP_FOREGROUND, SEP_AUTO, SEP_FRAME, SEP_CUSTOM };
 enum follow_mode { FOLLOW_NONE, FOLLOW_MOUSE, FOLLOW_KEYBOARD };
-enum markup_mode { MARKUP_NULL, MARKUP_NO, MARKUP_STRIP, MARKUP_FULL };
+enum mouse_action { MOUSE_NONE, MOUSE_DO_ACTION, MOUSE_CLOSE_CURRENT, MOUSE_CLOSE_ALL };
 
-typedef struct _settings {
+struct separator_color_data {
+        enum separator_color type;
+        char *sep_color;
+};
+
+struct geometry {
+        int x;
+        int y;
+        unsigned int w;
+        unsigned int h;
+        bool negative_x;
+        bool negative_y;
+        bool negative_width;
+        bool width_set;
+};
+
+struct settings {
         bool print_notifications;
         bool per_monitor_dpi;
         enum markup_mode markup;
         bool stack_duplicates;
         bool hide_duplicate_count;
         char *font;
-        char *normbgcolor;
-        char *normfgcolor;
-        char *normframecolor;
-        char *critbgcolor;
-        char *critfgcolor;
-        char *critframecolor;
-        char *lowbgcolor;
-        char *lowfgcolor;
-        char *lowframecolor;
+        struct notification_colors colors_low;
+        struct notification_colors colors_norm;
+        struct notification_colors colors_crit;
         char *format;
         gint64 timeouts[3];
         char *icons[3];
         unsigned int transparency;
-        char *geom;
+        struct geometry geometry;
         char *title;
         char *class;
         int shrink;
@@ -53,8 +65,7 @@ typedef struct _settings {
         int separator_height;
         int padding;
         int h_padding;
-        enum separator_color sep_color;
-        char *sep_custom_color_str;
+        struct separator_color_data sep_color;
         int frame_width;
         char *frame_color;
         int startup_notification;
@@ -62,19 +73,24 @@ typedef struct _settings {
         char *dmenu;
         char **dmenu_cmd;
         char *browser;
-        enum icon_position_t icon_position;
+        char **browser_cmd;
+        enum icon_position icon_position;
         int max_icon_size;
         char *icon_path;
         enum follow_mode f_mode;
         bool always_run_script;
-        keyboard_shortcut close_ks;
-        keyboard_shortcut close_all_ks;
-        keyboard_shortcut history_ks;
-        keyboard_shortcut context_ks;
+        struct keyboard_shortcut close_ks;
+        struct keyboard_shortcut close_all_ks;
+        struct keyboard_shortcut history_ks;
+        struct keyboard_shortcut context_ks;
         bool force_xinerama;
-} settings_t;
+        int corner_radius;
+        enum mouse_action mouse_left_click;
+        enum mouse_action mouse_middle_click;
+        enum mouse_action mouse_right_click;
+};
 
-extern settings_t settings;
+extern struct settings settings;
 
 void load_settings(char *cmdline_config_path);
 
