@@ -2,22 +2,32 @@
 #include "utils.h"
 
 #include <assert.h>
+#include <ctype.h>
 #include <errno.h>
 #include <glib.h>
 #include <stdio.h>
 #include <stdlib.h>
-#include <string.h>
+#include <time.h>
 
+#include "log.h"
+
+/* see utils.h */
 char *string_replace_char(char needle, char replacement, char *haystack)
 {
+        ASSERT_OR_RET(haystack, NULL);
+
         char *current = haystack;
-        while ((current = strchr(current, needle)) != NULL)
+        while ((current = strchr(current, needle)))
                 *current++ = replacement;
         return haystack;
 }
 
+/* see utils.h */
 char *string_replace_at(char *buf, int pos, int len, const char *repl)
 {
+        assert(buf);
+        assert(repl);
+
         char *tmp;
         int size, buf_len, repl_len;
 
@@ -42,19 +52,13 @@ char *string_replace_at(char *buf, int pos, int len, const char *repl)
         return tmp;
 }
 
-char *string_replace(const char *needle, const char *replacement, char *haystack)
-{
-        char *start;
-        start = strstr(haystack, needle);
-        if (start == NULL) {
-                return haystack;
-        }
-
-        return string_replace_at(haystack, (start - haystack), strlen(needle), replacement);
-}
-
+/* see utils.h */
 char *string_replace_all(const char *needle, const char *replacement, char *haystack)
 {
+        ASSERT_OR_RET(haystack, NULL);
+        assert(needle);
+        assert(replacement);
+
         char *start;
         int needle_pos;
         int needle_len, repl_len;
@@ -67,7 +71,7 @@ char *string_replace_all(const char *needle, const char *replacement, char *hays
         start = strstr(haystack, needle);
         repl_len = strlen(replacement);
 
-        while (start != NULL) {
+        while (start) {
                 needle_pos = start - haystack;
                 haystack = string_replace_at(haystack, needle_pos, needle_len, replacement);
                 start = strstr(haystack + needle_pos + repl_len, needle);
@@ -75,13 +79,14 @@ char *string_replace_all(const char *needle, const char *replacement, char *hays
         return haystack;
 }
 
+/* see utils.h */
 char *string_append(char *a, const char *b, const char *sep)
 {
-        if (!a || *a == '\0') {
+        if (STR_EMPTY(a)) {
                 g_free(a);
                 return g_strdup(b);
         }
-        if (!b || *b == '\0')
+        if (STR_EMPTY(b))
                 return a;
 
         char *new;
@@ -92,11 +97,29 @@ char *string_append(char *a, const char *b, const char *sep)
         g_free(a);
 
         return new;
-
 }
 
+/* see utils.h */
+char *string_strip_quotes(const char *value)
+{
+        ASSERT_OR_RET(value, NULL);
+
+        size_t len = strlen(value);
+        char *s;
+
+        if (value[0] == '"' && value[len-1] == '"')
+                s = g_strndup(value + 1, len-2);
+        else
+                s = g_strdup(value);
+
+        return s;
+}
+
+/* see utils.h */
 void string_strip_delimited(char *str, char a, char b)
 {
+        assert(str);
+
         int iread=-1, iwrite=0, copen=0;
         while (str[++iread] != 0) {
                 if (str[iread] == a) {
@@ -110,13 +133,14 @@ void string_strip_delimited(char *str, char a, char b)
         str[iwrite] = 0;
 }
 
+/* see utils.h */
 char *string_to_path(char *string)
 {
 
-        if (string && 0 == strncmp(string, "~/", 2)) {
+        if (string && STRN_EQ(string, "~/", 2)) {
                 char *home = g_strconcat(getenv("HOME"), "/", NULL);
 
-                string = string_replace("~/", home, string);
+                string = string_replace_at(string, 0, 2, home);
 
                 g_free(home);
         }
@@ -124,9 +148,9 @@ char *string_to_path(char *string)
         return string;
 }
 
+/* see utils.h */
 gint64 string_to_time(const char *string)
 {
-
         assert(string);
 
         errno = 0;
@@ -134,40 +158,50 @@ gint64 string_to_time(const char *string)
         gint64 val = strtoll(string, &endptr, 10);
 
         if (errno != 0) {
-                fprintf(stderr, "ERROR: Time: '%s': %s.\n", string, strerror(errno));
+                LOG_W("Time: '%s': %s.", string, strerror(errno));
                 return 0;
         } else if (string == endptr) {
-                fprintf(stderr, "ERROR: Time: No digits found.\n");
+                LOG_W("Time: '%s': No digits found.", string);
                 return 0;
         } else if (errno != 0 && val == 0) {
-                fprintf(stderr, "ERROR: Time: '%s' unknown error.\n", string);
+                LOG_W("Time: '%s': Unknown error.", string);
                 return 0;
         } else if (errno == 0 && !*endptr) {
-                return val * G_USEC_PER_SEC;
+                return S2US(val);
         }
 
         // endptr may point to a separating space
-        while (*endptr == ' ')
+        while (isspace(*endptr))
                 endptr++;
 
-        if (0 == strncmp(endptr, "ms", 2))
+        if (STRN_EQ(endptr, "ms", 2))
                 return val * 1000;
-        else if (0 == strncmp(endptr, "s", 1))
-                return val * G_USEC_PER_SEC;
-        else if (0 == strncmp(endptr, "m", 1))
-                return val * G_USEC_PER_SEC * 60;
-        else if (0 == strncmp(endptr, "h", 1))
-                return val * G_USEC_PER_SEC * 60 * 60;
-        else if (0 == strncmp(endptr, "d", 1))
-                return val * G_USEC_PER_SEC * 60 * 60 * 24;
+        else if (STRN_EQ(endptr, "s", 1))
+                return S2US(val);
+        else if (STRN_EQ(endptr, "m", 1))
+                return S2US(val) * 60;
+        else if (STRN_EQ(endptr, "h", 1))
+                return S2US(val) * 60 * 60;
+        else if (STRN_EQ(endptr, "d", 1))
+                return S2US(val) * 60 * 60 * 24;
         else
                 return 0;
 }
 
-void die(char *text, int exit_value)
+/* see utils.h */
+gint64 time_monotonic_now(void)
 {
-        fputs(text, stderr);
-        exit(exit_value);
-}
+        struct timespec tv_now;
 
+        /* On Linux, BOOTTIME is the correct monotonic time,
+         * as BOOTTIME counts onwards during sleep. For all other
+         * POSIX compliant OSes, MONOTONIC should also count onwards
+         * during system sleep. */
+#ifdef __linux__
+        clock_gettime(CLOCK_BOOTTIME, &tv_now);
+#else
+        clock_gettime(CLOCK_MONOTONIC, &tv_now);
+#endif
+        return S2US(tv_now.tv_sec) + tv_now.tv_nsec / 1000;
+}
 /* vim: set tabstop=8 shiftwidth=8 expandtab textwidth=0: */
