@@ -1,9 +1,7 @@
+#include "../src/option_parser.c"
 #include "greatest.h"
 
-#include <stdbool.h>
-#include <glib.h>
-
-#include "src/option_parser.h"
+extern const char *base;
 
 TEST test_next_section(void)
 {
@@ -53,6 +51,16 @@ TEST test_ini_get_string(void)
         free(ptr);
         ASSERT_STR_EQ("A string \"with quotes\"", (ptr = ini_get_string(string_section, "quoted_with_quotes", "")));
         free(ptr);
+        ASSERT_STR_EQ("A\" string with quotes\"", (ptr = ini_get_string(string_section, "unquoted_with_quotes", "")));
+        free(ptr);
+
+        ASSERT_STR_EQ("String with a", (ptr = ini_get_string(string_section, "quoted_comment", "")));
+        free(ptr);
+        ASSERT_STR_EQ("String with a", (ptr = ini_get_string(string_section, "unquoted_comment", "")));
+        free(ptr);
+        ASSERT_STR_EQ("#ffffff", (ptr = ini_get_string(string_section, "color_comment", "")));
+        free(ptr);
+
 
         ASSERT_STR_EQ("default value", (ptr = ini_get_string(string_section, "nonexistent", "default value")));
         free(ptr);
@@ -97,6 +105,10 @@ TEST test_ini_get_int(void)
 
 TEST test_ini_get_double(void)
 {
+        if (2.3 != atof("2.3")) {
+                SKIPm("Skipping test_ini_get_double, as it seems we're running under musl+valgrind!");
+        }
+
         char *double_section = "double";
         ASSERT_EQ(1, ini_get_double(double_section, "simple", 0));
         ASSERT_EQ(1.5, ini_get_double(double_section, "decimal", 0));
@@ -152,6 +164,10 @@ TEST test_cmdline_get_int(void)
 
 TEST test_cmdline_get_double(void)
 {
+        if (2.3 != atof("2.3")) {
+                SKIPm("Skipping test_cmdline_get_double, as it seems we're running under musl+valgrind!");
+        }
+
         ASSERT_EQ(2, cmdline_get_double("-simple_double", 0, ""));
         ASSERT_EQ(5.2, cmdline_get_double("-double", 0, ""));
         ASSERT_EQ(3.14, cmdline_get_double("-nonexistent", 3.14, ""));
@@ -258,6 +274,10 @@ TEST test_option_get_int(void)
 
 TEST test_option_get_double(void)
 {
+        if (2.3 != atof("2.3")) {
+                SKIPm("Skipping test_option_get_double, as it seems we're running under musl+valgrind!");
+        }
+
         char *double_section = "double";
         ASSERT_EQ(2, option_get_double(double_section, "simple", "-simple_double", 0, ""));
         ASSERT_EQ(5.2, option_get_double(double_section, "simple", "-double", 0, ""));
@@ -278,8 +298,9 @@ TEST test_option_get_bool(void)
 
 SUITE(suite_option_parser)
 {
-        FILE *config_file = fopen("data/test-ini", "r");
-        if (config_file == NULL) {
+        char *config_path = g_strconcat(base, "/data/test-ini", NULL);
+        FILE *config_file = fopen(config_path, "r");
+        if (!config_file) {
                 fputs("\nTest config file 'data/test-ini' couldn't be opened, failing.\n", stderr);
                 exit(1);
         }
@@ -312,6 +333,8 @@ SUITE(suite_option_parser)
         RUN_TEST(test_option_get_int);
         RUN_TEST(test_option_get_double);
         RUN_TEST(test_option_get_bool);
+
+        g_free(config_path);
         free_ini();
         g_strfreev(argv);
         fclose(config_file);
